import numpy as np
from geostat import gp, Model, NormalizingFeaturizer
import geostat.metric as gm

def test_euclidean():
    # Create 100 random locations in a square centered on the origin.
    locs = np.random.uniform(-1., 1., [1000, 3])

    metric = gm.Euclidean(scale=[1., 1., 'zscale'])
    covariance = \
        gp.GammaExponential(metric=metric) + \
        gp.SquaredExponential(range=2., sill='alpha') + \
        gp.Noise()

    # Generating GP.
    model1 = Model(
        latent = covariance,
        parameters = dict(alpha=1., zscale=5., range=0.5, sill=1., gamma=1., nugget=1.),
        verbose=True)

    # Generate data.
    vals = model1.generate(locs).vals

    # Fit GP.
    model2 = Model(
        latent = covariance,
        parameters = dict(alpha=2., zscale=1., range=1.0, sill=0.5, gamma=0.5, nugget=0.5),
        verbose=True).fit(locs, vals, iters=500)

    # Interpolate using GP.
    N = 10
    xx, yy, zz = np.meshgrid(np.linspace(-1, 1, N), np.linspace(-1, 1, N), np.linspace(-1, 1, N))
    locs2 = np.stack([xx, yy, zz], axis=-1).reshape([-1, 3])

    mean, var = model2.predict(locs2)
    mean2, var2 = model2.predict(locs2)

    assert np.all(mean == mean2)
    assert np.all(var == var2)

def test_poincare():
    np.random.seed(1235)

    # Create random locations in a square centered on the origin.
    locs = np.random.uniform(-2., 2., [1000, 3])
    locs[:, 2] = 3. - np.exp(locs[:, 2]) # Make z to be 3 or lower.

    # Transform z to be positive, and make it the first axis.
    zmax = locs[:, 2].max()
    def xform(x, y, z): return zmax-z, x, y

    metric = gm.Poincare(xform=xform, scale=['zscale', 1., 1.], zoff='zoff')
    covariance = \
        gp.GammaExponential(metric=metric) + \
        gp.SquaredExponential(metric=metric, range=2., sill='alpha') + \
        gp.Noise()

    # Generating GP.
    model1 = Model(
        latent = covariance,
        parameters = dict(zoff=2.0, alpha=1., zscale=5., range=0.5, sill=1., gamma=1., nugget=1.),
        verbose=True)

    # Generate data.
    vals = model1.generate(locs).vals

    # Fit GP.
    model2 = Model(
        latent = covariance,
        parameters = dict(zoff=2.0, alpha=2., zscale=1., range=1.0, sill=0.5, gamma=0.5, nugget=0.5),
        verbose=True).fit(locs, vals, iters=500)

    # Interpolate using GP.
    N = 10
    xx, yy, zz = np.meshgrid(np.linspace(-1, 1, N), np.linspace(-1, 1, N), np.linspace(-1, 1, N))
    locs2 = np.stack([xx, yy, zz], axis=-1).reshape([-1, 3])

    mean, var = model2.predict(locs2)
    mean2, var2 = model2.predict(locs2)

    assert np.all(mean == mean2)
    assert np.all(var == var2)




from dataclasses import dataclass
from typing import Callable, Dict

import numpy as np

# Tensorflow is extraordinarily noisy. Catch warnings during import.
import warnings
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    import tensorflow as tf

from .op import Op
from .param import get_parameter_values, ppp, upp, bpp

def ed(x, a=-1):
    return tf.expand_dims(x, a)

class PerAxisDist2(Op):
    def __init__(self):
        super().__init__({}, dict(locs1='locs1', locs2='locs2'))

    def __call__(self, p, e):
        x1 = e['locs1']
        x2 = e['locs2']
        return tf.square(ed(x1, 1) - ed(x2, 0))

class Metric(Op):
    pass

def get_scale_vars(scale):
    if scale is not None:
        return [p for s in scale for p in ppp(s)]
    else:
        return []

class Euclidean(Metric):
    def __init__(self, scale=None):
        fa = dict(scale=scale)
        super().__init__(fa, dict(pa_d2='per_axis_dist2'))

    def vars(self):
        return get_scale_vars(self.fa['scale'])

    def __call__(self, p, e):
        v = get_parameter_values(self.fa, p)
        if v['scale'] is not None:
            return tf.einsum('abc,c->ab', e['pa_d2'], tf.square(v['scale']))
        else:
            return tf.reduce_sum(e['pa_d2'], axis=-1)

class Poincare(Metric):
    def __init__(self, xform: Callable, zoff='zoff', scale=None):
        fa = dict(zoff=zoff, scale=scale)
        self.xform = xform
        super().__init__({}, dict(locs1='locs1', locs2='locs2'))

    def vars(self):
        return ppp(self.fa['zoff']) + get_scale_vars(self.fa['scale'])

    def __call__(self, p, e):
        v = get_parameter_values(self.fa, p)

        xlocs1 = tf.stack(self.xform(*tf.unstack(e['locs1'], axis=1)), axis=1)
        xlocs2 = tf.stack(self.xform(*tf.unstack(e['locs2'], axis=1)), axis=1)
        zoff = v['zoff']

        # Maybe scale locations and zoff.
        if v['scale'] is not None:
            xlocs *= v['scale']
            zoff *= v['scale'][0]

        z = xlocs[:, 0] + zoff
        zz = z * ed(z, -1)

        d2 = tf.reduce_sum(tf.square(ed(xlocs1, 0) - ed(xlocs2, 1)), axis=-1)
        d2 = tf.asinh(0.5 * tf.sqrt(d2 / zz))
        d2 = tf.square(2.0 * zoff * d2)

        return d2

def is_distance_matrix(m):
    assert len(m.shape) == 2
    assert m.shape[0] == m.shape[1]

    a = np.ones_like(m) * float('inf')
    for i in range(m.shape[0]):
        a = np.minimum(a, m[i:i+1, :] + m[:, i:i+1])

    # print(np.nonzero(~np.equal(a, m)))
    return np.allclose(a, m)



import numpy as np
from geostat import gp, Model, Featurizer, NormalizingFeaturizer, Observation
import numpy as np
import tensorflow as tf
from argparse import Namespace


def test_delta():
    # Create 100 random locations in a square centered on the origin.
    np.random.seed(123)
    tf.random.set_seed(123)

    counts = np.array([100, 150, 200, 200]) * 3
    locs1 = np.random.uniform(-1., 1., [sum(counts) // 3, 4]) * [10., 10., 1., 1.]
    locs1 = locs1[:, np.newaxis, :] + np.pad(np.linspace(0.01, 0.1, 3)[:, np.newaxis], ((0, 0), (2, 1)))
    locs1 = locs1.reshape([-1, 4])

    cats1 = [x for i, count in enumerate(counts) for x in [i] * count]
    # np.random.shuffle(cats1)
    
    # print(locs1.shape)
    # for i, (row, cat) in enumerate(zip(locs1, cats1)):
    #     print(i, ':', cat, ':', locs1[i])

    # Initialize featurizer of location for trends.
    def trend_terms(x, y, z, t): return z, z*z, z*z*z
    featurizer = NormalizingFeaturizer(trend_terms, locs1)

    covu = gp.TrendPrior(featurizer, alpha='au') \
         + gp.SquaredExponential(sill='su1', range='ru1', scale=[1., 1., 'zu', 'tu'])
    covp = gp.TrendPrior(featurizer, alpha='ap') \
         + gp.SquaredExponential(sill='sp1', range='rp1', scale=[1., 1., 'zp', 0.  ])
    covt = gp.TrendPrior(featurizer, alpha='at') \
         + gp.SquaredExponential(sill='st1', range='rt1', scale=[1., 1., 'zt', 'tt'])

    feat_r = NormalizingFeaturizer(lambda x, y, z, t: (), locs1)

    def neg_transformed_natural_gradient(x, y, z, t):
        degF = 62.2 + 57.1 * (0.15 - z)
        return -(tf.math.log(degF + 6.77) - tf.math.log(75 + 6.77))

    obsu = Observation([1., 0., 0.], gp.Noise(nugget='nu') + gp.Delta(dsill='wu', axes=[0, 1]))
    obsp = Observation([0., 1., 0.], gp.Noise(nugget='np') + gp.Delta(dsill='wp', axes=[0, 1]))
    obst = Observation([0., 0., 1.], gp.Noise(nugget='nt') + gp.Delta(dsill='wt', axes=[0, 1]))
    obsR = Observation(['cu', 'cp', 0.],
        gp.Trend(Featurizer(neg_transformed_natural_gradient), beta=[1.]) +
        gp.Noise(nugget='nr') + gp.Delta(dsill='wr', axes=[0, 1]) + gp.TrendPrior(feat_r, alpha='ar'))

    p_init = {'au': 1.,
     'su1': 0.01,
     'ru1': 0.5,
     'cu': -0.5,
     'wu': 0.01,
     'nu': 0.02,
     'ap': 1.,
     'sp1': 0.04,
     'rp1': 0.4,
     'cp': -0.3,
     'wp': 0.002,
     'np': 0.001,
     'at': 0.1,
     'st1': 0.02,
     'rt1': 1.,
     'wt': 0.01,
     'nt': 0.01,
     'ar': 1.,
     'wr': 0.03,
     'nr': 0.001,
     'zu': 10.,
     'zp': 15.,
     'zt': 20.,
     'tu': 10.,
     'tt': 10.}

    def report(p):
        p = {k: (v.numpy() if hasattr(v, 'numpy') else v) for k, v in p.items()}
        p = Namespace(**p)
        print(f'[iter {p.iter:5d}] [ll {p.ll:.3f}] [reg {p.reg:.3f}] [time {p.time:.1f}]')
        print(f'   {"alp":>6s} {" zs":>6s} {" ts":>6s} {" sil":>6s} {" rng":>6s} {"  c":>6s} {" wi":>6s} {"nug":>6s}')
        print(f'u: {p.au:6.3f} {p.zu:6.3f} {p.tu:6.3f} {p.su1:6.3f} {p.ru1:6.3f} {p.cu:6.3f} {p.wu:6.3f} {p.nu:6.3f}')
        print(f'p: {p.ap:6.3f} {p.zp:6.3f} {"   ":>6s} {p.sp1:6.3f} {p.rp1:6.3f} {p.cp:6.3f} {p.wp:6.3f} {p.np:6.3f}')
        print(f't: {p.at:6.3f} {p.zt:6.3f} {p.tt:6.3f} {p.st1:6.3f} {p.rt1:6.3f} {-1. :6.3f} {p.wt:6.3f} {p.nt:6.3f}')
        print(f'r: {p.ar:6.3f} {"    ":6s} {"    ":6s} {"     ":6s} {"     ":6s} {"    ":6s} {p.wr:6.3f} {p.nr:6.3f}')

    model1 = Model(
        latent = [covu, covp, covt],
        observed = [obsu, obsp, obst, obsR],
        parameters = p_init,
        report = report,
        verbose = True)

    # Generate data.
    vals1 = model1.generate(locs1, cats1).vals

    model2 = Model(
        latent = [covu, covp, covt],
        observed = [obsu, obsp, obst, obsR],
        parameters = {k: 2*v if 'g' not in k else v for k, v in p_init.items()},
        report = report,
        verbose=True).fit(locs1, vals1, cats1, reg=1., iters=500)



import numpy as np
from geostat import gp, Model, Featurizer, NormalizingFeaturizer

def test_multigp():

    # Create random locations in a square centered on the origin.
    N = 200
    locs1 = np.random.uniform(-1., 1., [3*N, 2])
    # Triple data with offsets.
    N *= 3
    locs1 = np.concatenate([locs1 - [0.1, 0], locs1, locs1 + [0.1, 0]], axis=-1).reshape([-1, 2])

    # Initialize featurizer of location for trends.
    def trend_terms(x, y): return x, y, x*y
    featurizer = NormalizingFeaturizer(trend_terms, locs1)
    cov1 = gp.TrendPrior(featurizer, alpha='a1') + gp.SquaredExponential(sill='s1', range='r1')
    cov2 = gp.TrendPrior(featurizer, alpha='a2') + gp.SquaredExponential(sill='s2', range='r2')

    f2 = Featurizer(lambda x, y: (1, x + y*y))
    obs1 = gp.Observation([1., 0.], gp.Trend(f2, beta=[0., 0.]) + gp.Noise(nugget='n1'))
    obs2 = gp.Observation([0., 1.], gp.Trend(f2, beta=['c2', 0.]) + gp.Noise(nugget='n2'))
    obs3 = gp.Observation(['k1', 'k2'], gp.Trend(f2, beta=[0., 1.]) + gp.Noise(nugget='n3') + gp.Delta(dsill='d', axes=[1]))

    # Generating GP.
    model1 = Model(
        latent = [cov1, cov2],
        observed = [obs1, obs2, obs3],
        parameters = dict(
            a1=1., s1=1., r1=0.5, k1=2.,
            a2=1., s2=1., r2=0.5, k2=3., c2=1.,
            n1=0.1, n2=0.2, n3=0.3, d=0.1),
        verbose=True)

    # Generate data.
    cats1 = [0] * N + [1] * N + [2] * N
    vals1 = model1.generate(locs1, cats1).vals

    # Fit GP.
    model2 = Model(
        latent = [cov1, cov2],
        observed = [obs1, obs2, obs3],

        parameters = dict(
            a1=1., s1=1., r1=1., k1=0.,
            a2=1., s2=1., r2=1., k2=0., c2=0.,
            n1=0.1, n2=0.1, n3=0.1, d=0.1),
        verbose=True).fit(locs1, vals1, cats1, iters=2000)

    # Interpolate using GP.
    N = 20
    xx, yy = np.meshgrid(np.linspace(-1, 1, N), np.linspace(-1, 1, N))
    locs2 = np.stack([xx, yy], axis=-1)
    for cats in range(3):
        mean, var = model2.predict(locs2, cats * np.ones_like(locs2[..., 0], dtype=np.int32))



import time
from collections import defaultdict
from dataclasses import dataclass, replace
from typing import Callable, Dict, List, Optional, Union
import numpy as np
from scipy.special import expit, logit
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '1'

# Tensorflow is extraordinarily noisy. Catch warnings during import.
import warnings
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    import tensorflow as tf
    import tensorflow_probability as tfp
    tfd = tfp.distributions
    from tensorflow.linalg import LinearOperatorFullMatrix as LOFullMatrix
    from tensorflow.linalg import LinearOperatorBlockDiag as LOBlockDiag

import tensorflow_probability as tfp

from . import gp
from .op import Op
from .param import PaperParameter, ParameterSpace, Bound
from .metric import Euclidean, PerAxisDist2
from .param import get_parameter_values, ppp, upp, bpp

MVN = tfp.distributions.MultivariateNormalTriL

__all__ = ['Model', 'Featurizer', 'NormalizingFeaturizer']

class NormalizingFeaturizer:
    """
    Produces featurized locations (F matrix) and remembers normalization
    parameters.  Adds an intercept feature.
    """
    def __init__(self, featurization, locs):
        self.unnorm_featurizer = Featurizer(featurization)
        F_unnorm = self.unnorm_featurizer(locs)
        self.unnorm_mean = tf.reduce_mean(F_unnorm, axis=0)
        self.unnorm_std = tf.math.reduce_std(F_unnorm, axis=0)

    def __call__(self, locs):
        ones = tf.ones([tf.shape(locs)[0], 1], dtype=tf.float32)
        F_unnorm = self.unnorm_featurizer(locs)
        F_norm = (F_unnorm - self.unnorm_mean) / self.unnorm_std
        return tf.concat([ones, F_norm], axis=1)

class Featurizer:
    """
    Produces featurized locations (F matrix).
    """
    def __init__(self, featurization):
        self.featurization = featurization

    def __call__(self, locs):
        locs = tf.cast(locs, tf.float32)
        if self.featurization is None: # No features.
            return tf.ones([tf.shape(locs)[0], 0], dtype=tf.float32)

        feats = self.featurization(*tf.unstack(locs, axis=1))
        if isinstance(feats, tuple): # One or many features.
            if len(feats) == 0:
                return tf.ones([tf.shape(locs)[0], 0], dtype=tf.float32)
            else:
                feats = self.featurization(*tf.unstack(locs, axis=1))
                feats = [tf.broadcast_to(tf.cast(f, tf.float32), [tf.shape(locs)[0]]) for f in feats]
                return tf.stack(feats, axis=1)
        else: # One feature.
            return e(feats)

def e(x, a=-1):
    return tf.expand_dims(x, a)

def block_diag(blocks):
    """Return a dense block-diagonal matrix."""
    return LOBlockDiag([LOFullMatrix(b) for b in blocks]).to_dense()

@tf.function
def gp_covariance(covariance, observation, locs, cats, p):
    return gp_covariance2(covariance, observation, locs, cats, locs, cats, 0, p)

@tf.function
def gp_covariance2(covariance, observation, locs1, cats1, locs2, cats2, offset, p):
    """
    `offset` is i2-i1, where i1 and i2 are the starting indices of locs1
    and locs2.  It is used to create the diagonal non-zero elements
    of a Noise covariance function.  An non-zero offset results in a
    covariance matrix with non-zero entries along an off-center diagonal.
    """

    # assert np.all(cats1 == np.sort(cats1)), '`cats1` must be in non-descending order'
    # assert np.all(cats2 == np.sort(cats2)), '`cats2` must be in non-descending order'

    locs1 = tf.cast(locs1, tf.float32)
    locs2 = tf.cast(locs2, tf.float32)

    cache = {}
    cache['offset'] = offset
    cache['locs1'] = locs1
    cache['locs2'] = locs2
    cache['per_axis_dist2'] = PerAxisDist2().run(cache, p)
    cache['euclidean'] = Euclidean().run(cache, p)

    MM, CC = zip(*[c.run(cache, p) for c in covariance])
    M = tf.stack(MM, axis=-1) # [locs, hidden].
    C = tf.stack(CC, axis=-1) # [locs, locs, hidden].

    numobs = len(observation)

    if numobs == 0:
        assert len(covariance) == 1, 'With no observation model, I only want one covariance model'
        C = tf.cast(C[..., 0], tf.float64)
        M = tf.cast(M[..., 0], tf.float64)
        return M, C

    A = tf.convert_to_tensor(gp.get_parameter_values([o.coefs for o in observation], p)) # [surface, hidden].
    M = tf.gather(tf.einsum('lh,sh->ls', M, A), cats1, batch_dims=1) # [locs]

    Aaug1 = tf.gather(A, cats1) # [locs, hidden].
    Aaug2 = tf.gather(A, cats2) # [locs, hidden].
    outer = tf.einsum('ac,bc->abc', Aaug1, Aaug2) # [locs, locs, hidden].
    C = tf.einsum('abc,abc->ab', C, outer) # [locs, locs].

    catcounts1 = tf.math.bincount(cats1, minlength=numobs, maxlength=numobs)
    catcounts2 = tf.math.bincount(cats2, minlength=numobs, maxlength=numobs)
    catindices1 = tf.math.cumsum(catcounts1, exclusive=True)
    catindices2 = tf.math.cumsum(catcounts2, exclusive=True)
    catdiffs = tf.unstack(catindices2 - catindices1, num=numobs)
    locsegs1 = tf.split(locs1, catcounts1, num=numobs)
    locsegs2 = tf.split(locs2, catcounts2, num=numobs)

    CC = [] # Observation noise submatrices.
    MM = [] # Mean subvectors.
    for sublocs1, sublocs2, catdiff, o in zip(locsegs1, locsegs2, catdiffs, observation):
        cache['offset'] = offset + catdiff
        cache['locs1'] = sublocs1
        cache['locs2'] = sublocs2
        cache['per_axis_dist2'] = PerAxisDist2().run(cache, p)
        cache['euclidean'] = Euclidean().run(cache, p)
        Msub, Csub = o.noise.run(cache, p)
        CC.append(Csub)
        MM.append(Msub)

    C += block_diag(CC)
    C = tf.cast(C, tf.float64)

    M += tf.concat(MM, axis=0)
    M = tf.cast(M, tf.float64)

    return M, C

@tf.function
def mvn_log_pdf(u, m, cov):
    """Log PDF of a multivariate gaussian."""
    u_adj = u - m
    logdet = tf.linalg.logdet(2 * np.pi * cov)
    quad = tf.matmul(e(u_adj, 0), tf.linalg.solve(cov, e(u_adj, -1)))[0, 0]
    return tf.cast(-0.5 * (logdet + quad), tf.float32)

@tf.function
def gp_log_likelihood(data, surf_params, covariance, observation):
    m, S = gp_covariance(covariance, observation, data['locs'], data['cats'], surf_params)
    u = tf.cast(data['vals'], tf.float64)
    return mvn_log_pdf(u, m, S)

def gp_train_step(optimizer, data, parameters, parameter_space, covariance, observation, reg=None):
    with tf.GradientTape() as tape:
        sp = parameter_space.get_surface(parameters)

        ll = gp_log_likelihood(data, sp, covariance, observation)

        if reg:
            reg_penalty = reg * tf.reduce_sum([c.reg(sp) for c in covariance])
        else:
            reg_penalty = 0.

        loss = -ll + reg_penalty

    gradients = tape.gradient(loss, parameters.values())
    optimizer.apply_gradients(zip(gradients, parameters.values()))
    return sp, ll, reg_penalty

def check_parameters(pps: List[PaperParameter], values: Dict[str, float]) -> Dict[str, Bound]:
    d = defaultdict(list)
    for pp in pps:
        d[pp.name].append(pp)
    out = {}
    for name, pps in d.items():
        lo = np.max([pp.lo for pp in pps])
        hi = np.min([pp.hi for pp in pps])
        assert lo < hi, 'Conflicting bounds for parameter `%s`' % name
        assert name in values, 'Parameter `%s` is missing' % name
        assert np.all(lo <= values[name]) and np.all(values[name] <= hi), 'Parameter `%s` is out of bounds' % name
        out[name] = Bound(lo, hi)
    return out

@dataclass
class Model():
    latent: List[gp.GP] = None
    observed: List[gp.Observation] = None
    parameters: Dict[str, np.ndarray] = None
    parameter_sample_size: Optional[int] = None
    locs: np.ndarray = None
    vals: np.ndarray = None
    cats: np.ndarray = None
    report: Callable = None
    verbose: bool = True

    def __post_init__(self):

        '''
        Parameters:
                x : Pandas DataFrame with columns for locations.

                u : A Pandas Series containing observations.

                featurization : function, optional
                    Should be a function that takes x1 (n-dim array of input data)
                    and returns the coordinates, i.e., x, y, x**2, y**2.
                    Example: def featurization(x1):
                                return x1[:, 0], x1[:, 1], x1[:, 0]**2, x1[:, 1]**2.
                    Default is None.

                latent : List[gp.GP]
                     Name of the covariance function to use in the GP.
                     Should be 'squared-exp' or 'gamma-exp'.
                     Default is 'squared-exp'.

                parameters : dict, optional
                    The starting point for the parameters.
                    Example: parameters=dict(range=2.0, sill=5.0, nugget=1.0).
                    Default is None.

                verbose : boolean, optional
                    Whether or not to print parameters.
                    Default is True.

        Performs Gaussian process training and prediction.
        '''

        super().__init__()

        assert self.latent is not None, 'I need at least one latent variable'
        if isinstance(self.latent, gp.GP):
            self.latent = [self.latent]

        if self.observed is None: self.observed = []

        # Default reporting function.
        def default_report(p, prefix=None):
            if prefix: print(prefix, end=' ')

            def fmt(x):
                if isinstance(x, tf.Tensor):
                    x = x.numpy()

                if isinstance(x, (int, np.int32, np.int64)):
                    return '{:5d}'.format(x)
                if isinstance(x, (float, np.float32, np.float64)):
                    return '{:5.2f}'.format(x)
                else:
                    with np.printoptions(precision=2, formatter={'floatkind': '{:5.2f}'.format}):
                        return str(x)

            print('[%s]' % (' '.join('%s %s' % (k, fmt(v)) for k, v in p.items())))

        if self.report == None: self.report = default_report

        if self.locs is not None: self.locs = np.array(self.locs)
        if self.vals is not None: self.vals = np.array(self.vals)
        if self.cats is not None: self.cats = np.array(self.cats)

        # Collect paraameters.
        if self.parameters is None: self.parameters = {}
        vv = {v for c in self.latent for v in c.gather_vars()}
        vv |= {v for o in self.observed for v in o.gather_vars()}

        self.parameter_space = ParameterSpace(check_parameters(vv, self.parameters))

    def fit(self, locs, vals, cats=None,
        step_size=0.01, iters=100, reg=None):

        # Permute datapoints if cats is given.
        if cats is not None:
            cats = np.array(cats)
            perm = np.argsort(cats)
            locs, vals, cats = locs[perm], vals[perm], cats[perm]

        # Data dict.
        self.data = {
            'locs': tf.constant(locs, dtype=tf.float32),
            'vals': tf.constant(vals, dtype=tf.float32),
            'cats': None if cats is None else tf.constant(cats, dtype=tf.int32)}

        up = self.parameter_space.get_underlying(self.parameters)

        optimizer = tf.keras.optimizers.Adam(learning_rate=step_size)

        j = 0 # Iteration count.
        for i in range(10):
            t0 = time.time()
            while j < (i + 1) * iters / 10:
                p, ll, reg_penalty = gp_train_step(optimizer, self.data, up, self.parameter_space,
                    self.latent, self.observed, reg)
                j += 1

            time_elapsed = time.time() - t0
            if self.verbose == True:
                self.report(dict(iter=j, ll=ll, time=time_elapsed, reg=reg_penalty, **p))

        new_parameters = self.parameter_space.get_surface(up, numpy=True)

        # Restore order if things were permuted.
        if cats is not None:
            revperm = np.argsort(perm)
            locs, vals, cats = locs[revperm], vals[revperm], cats[revperm]

        return replace(self, parameters = new_parameters, locs=locs, vals=vals, cats=cats)

    def mcmc(self, locs, vals, cats=None,
            chains=4, step_size=0.1, move_prob=0.5,
            samples=1000, burnin=500, report_interval=100):

        assert samples % report_interval == 0, '`samples` must be a multiple of `report_interval`'
        assert burnin % report_interval == 0, '`burnin` must be a multiple of `report_interval`'

        # Permute datapoints if cats is given.
        if cats is not None:
            cats = np.array(cats)
            perm = np.argsort(cats)
            locs, vals, cats = locs[perm], vals[perm], cats[perm]

        # Data dict.
        self.data = {
            'locs': tf.constant(locs, dtype=tf.float32),
            'vals': tf.constant(vals, dtype=tf.float32),
            'cats': None if cats is None else tf.constant(cats, dtype=tf.int32)}

        # Initial MCMC state.
        initial_up = self.parameter_space.get_underlying(self.parameters)

        # Unnormalized log posterior distribution.
        def g(up):
            sp = self.parameter_space.get_surface(up)
            return gp_log_likelihood(self.data, sp, self.latent, self.observed)

        def f(*up_flat):
            up = tf.nest.pack_sequence_as(initial_up, up_flat)
            ll = tf.map_fn(g, up, fn_output_signature=tf.float32)
            # log_prior = -tf.reduce_sum(tf.math.log(1. + tf.square(up_flat)), axis=0)
            return ll # + log_prior

        # Run the chain for a burst.
        @tf.function
        def run_chain(current_state, final_results, kernel, iters):
            samples, results, final_results = tfp.mcmc.sample_chain(
                num_results=iters,
                current_state=current_state,
                kernel=kernel,
                return_final_kernel_results=True,
                trace_fn=lambda _, results: results)

            return samples, results, final_results
        
        def new_state_fn(scale, dtype):
          direction_dist = tfd.Normal(loc=dtype(0), scale=dtype(1))
          scale_dist = tfd.Exponential(rate=dtype(1/scale))
          pick_dist = tfd.Bernoulli(probs=move_prob)

          def _fn(state_parts, seed):
            next_state_parts = []
            part_seeds = tfp.random.split_seed(
                seed, n=len(state_parts), salt='rwmcauchy')
            for sp, ps in zip(state_parts, part_seeds):
                pick = tf.cast(pick_dist.sample(sample_shape=sp.shape, seed=ps), tf.float32)
                direction = direction_dist.sample(sample_shape=sp.shape, seed=ps)
                scale_val = scale_dist.sample(seed=ps)
                next_state_parts.append(sp + tf.einsum('a...,a->a...', pick * direction, scale_val))
            return next_state_parts
          return _fn

        inv_temps = 0.5**np.arange(chains, dtype=np.float32)

        def make_kernel_fn(target_log_prob_fn):
            return tfp.mcmc.RandomWalkMetropolis(
                target_log_prob_fn=target_log_prob_fn,
                new_state_fn=new_state_fn(scale=step_size / np.sqrt(inv_temps), dtype=np.float32))

        kernel = tfp.mcmc.ReplicaExchangeMC(
            target_log_prob_fn=f,
            inverse_temperatures=inv_temps,
            make_kernel_fn=make_kernel_fn)

        # Do bursts.
        current_state = tf.nest.flatten(initial_up)
        final_results = None
        acc_states = []
        num_bursts = (samples + burnin) // report_interval
        burnin_bursts = burnin // report_interval
        for i in range(num_bursts):
            is_burnin = i < burnin_bursts

            if self.verbose and (i == 0 or i == burnin_bursts):
                print('BURNIN\n' if is_burnin else '\nSAMPLING')
            
            t0 = time.time()
            states, results, final_results = run_chain(current_state, final_results, kernel, report_interval)

            if self.verbose == True:
                if not is_burnin: print()
                accept_rates = results.post_swap_replica_results.is_accepted.numpy().mean(axis=0)
                print('[iter {:4d}] [time {:.1f}] [accept rates {}]'.format(
                    ((i if is_burnin else i - burnin_bursts) + 1) * report_interval,
                    time.time() - t0,
                    ' '.join([f'{x:.2f}' for x in accept_rates.tolist()])))

            if not is_burnin:
                acc_states.append(tf.nest.map_structure(lambda x: x.numpy(), states))
                all_states = [np.concatenate(x, 0) for x in zip(*acc_states)]
                up = tf.nest.pack_sequence_as(initial_up, all_states)
                sp = self.parameter_space.get_surface(up, numpy=True) 

                # Reporting
                if self.verbose == True:
                    for p in [5, 50, 95]:
                        x = tf.nest.map_structure(lambda x: np.percentile(x, p, axis=0), sp)
                        self.report(x, prefix=f'{p:02d}%ile')

            current_state = [s[-1] for s in states]

        posterior = self.parameter_space.get_surface(up, numpy=True)

        # Restore order if things were permuted.
        if cats is not None:
            revperm = np.argsort(perm)
            locs, vals, cats = locs[revperm], vals[revperm], cats[revperm]

        return replace(self, 
            parameters=posterior,
            parameter_sample_size=samples,
            locs=locs, vals=vals, cats=cats)

    def generate(self, locs, cats=None):
        assert self.locs is None and self.vals is None, 'Conditional generation not yet supported'
        assert self.parameter_sample_size is None, 'Generation from a distribution not yet supported'

        locs = np.array(locs)

        # Permute datapoints if cats is given.
        if cats is not None:
            cats = np.array(cats)
            perm = np.argsort(cats)
            locs, cats = locs[perm], cats[perm]

        up = self.parameter_space.get_underlying(self.parameters)

        p = self.parameter_space.get_surface(up)

        m, S = gp_covariance(
            self.latent,
            self.observed,
            tf.constant(locs, dtype=tf.float32),
            None if cats is None else tf.constant(cats, dtype=tf.int32),
            p)

        vals = MVN(m, tf.linalg.cholesky(S)).sample().numpy()

        # Restore order if things were permuted.
        if cats is not None:
            revperm = np.argsort(perm)
            locs, vals, cats = locs[revperm], vals[revperm], cats[revperm]

        return replace(self, locs=locs, vals=vals, cats=cats)

    def predict(self, locs2, cats2=None, subsample=None, reduce=None, tracker=None):
        '''
        Performs GP predictions of the mean and variance.
        Has support for batch predictions for large data sets.
        '''

        assert subsample is None or self.parameter_sample_size is not None, \
            '`subsample` is only valid with sampled parameters'

        assert reduce is None or self.parameter_sample_size is not None, \
            '`reduce` is only valid with sampled parameters'

        assert subsample is None or reduce is None, \
            '`subsample` and `reduce` cannot both be given'

        if tracker is None: tracker = lambda x: x

        assert self.locs.shape[-1] == locs2.shape[-1], 'Mismatch in location dimentions'
        if cats2 is not None:
            assert cats2.shape == locs2.shape[:-1], 'Mismatched shapes in cats and locs'

        def interpolate_batch(A11i, locs1, vals1diff, cats1, locs2, cats2, parameters):

            N1 = len(locs1) # Number of measurements.

            # Permute datapoints if cats is given.
            if cats2 is not None:
                perm = np.argsort(cats2)
                locs2, cats2 = locs2[perm], cats2[perm]

            _, A12 = gp_covariance2(
                self.latent,
                self.observed,
                tf.constant(locs1, dtype=tf.float32),
                None if cats1 is None else tf.constant(cats1, dtype=tf.int32),
                tf.constant(locs2, dtype=tf.float32),
                None if cats2 is None else tf.constant(cats2, dtype=tf.int32),
                N1,
                parameters)

            m2, A22 = gp_covariance(
                self.latent,
                self.observed,
                tf.constant(locs2, dtype=tf.float32),
                None if cats2 is None else tf.constant(cats2, dtype=tf.int32),
                parameters)

            # Restore order if things were permuted.
            if cats2 is not None:
                revperm = np.argsort(perm)
                m2 = tf.gather(m2, revperm)
                A12 = tf.gather(A12, revperm, axis=-1)
                A22 = tf.gather(tf.gather(A22, revperm), revperm, axis=-1)

            u2_mean = m2 + tf.einsum('ab,a->b', A12, tf.einsum('ab,b->a', A11i, vals1diff))
            u2_var = tf.linalg.diag_part(A22) -  tf.einsum('ab,ab->b', A12, tf.matmul(A11i, A12))

            return u2_mean, u2_var

        def interpolate(locs1, vals1, cats1, locs2, cats2, parameters):
            # Interpolate in batches.
            batch_size = self.locs.shape[0] // 2

            for_gp = []
            locs2r = locs2.reshape([-1, locs2.shape[-1]])
            if cats2 is not None:
                cats2r = cats2.ravel()
            else:
                cats2r = np.zeros_like(locs2r[..., 0], np.int32)

            for start in np.arange(0, len(locs2r), batch_size):
                stop = start + batch_size
                subset = locs2r[start:stop], cats2r[start:stop]
                for_gp.append(subset)

            # Permute datapoints if cats is given.
            if cats1 is not None:
                perm = np.argsort(cats1)
                locs1, vals1, cats1 = locs1[perm], vals1[perm], cats1[perm]

            m1, A11 = gp_covariance(
                self.latent,
                self.observed,
                tf.constant(locs1, dtype=tf.float32),
                None if cats1 is None else tf.constant(cats1, dtype=tf.int32),
                parameters)

            A11i = tf.linalg.inv(A11)

            u2_mean_s = []
            u2_var_s = []

            for locs_subset, cats_subset in for_gp:
                u2_mean, u2_var = interpolate_batch(A11i, locs1, vals1 - m1, cats1, locs_subset, cats_subset, parameters)
                u2_mean = u2_mean.numpy()
                u2_var = u2_var.numpy()
                u2_mean_s.append(u2_mean)
                u2_var_s.append(u2_var)

            u2_mean = np.concatenate(u2_mean_s).reshape(locs2.shape[:-1])
            u2_var = np.concatenate(u2_var_s).reshape(locs2.shape[:-1])

            return u2_mean, u2_var

        if self.parameter_sample_size is None:
            m, v = interpolate(self.locs, self.vals, self.cats, locs2, cats2, self.parameters)
        elif reduce == 'median':
            p = tf.nest.map_structure(lambda x: np.quantile(x, 0.5, axis=0).astype(np.float32), self.parameters)
            m, v = interpolate(self.locs, self.vals, self.cats, locs2, cats2, p)
        elif reduce == 'mean':
            p = tf.nest.map_structure(lambda x: x.mean(axis=0).astype(np.float32), self.parameters)
            m, v = interpolate(self.locs, self.vals, self.cats, locs2, cats2, p)
        else:
            samples = self.parameter_sample_size

            if subsample is not None:
                assert subsample <= samples, '`subsample` may not exceed sample size'
            else:
                subsample = samples

            # Thin by picking roughly equally-spaced samples.
            a = np.arange(samples) * subsample / samples % 1
            pick = np.concatenate([[True], a[1:] >= a[:-1]])
            parameters = tf.nest.map_structure(lambda x: x[pick], self.parameters)

            # Make a prediction for each sample.
            results = []
            for i in tracker(range(subsample)):
                p = tf.nest.map_structure(lambda x: x[i], parameters)
                results.append(interpolate(self.locs, self.vals, self.cats, locs2, cats2, p))

            mm, vv = [np.stack(x) for x in zip(*results)]
            m = mm.mean(axis=0)
            v = (np.square(mm) + vv).mean(axis=0) - np.square(m)

        return m, v



from dataclasses import dataclass, replace
import numpy as np
import pandas as pd
from scipy.spatial import ConvexHull
from shapely.geometry import MultiPoint, Point, Polygon, shape

__all__ = ['Mesh']

@dataclass
class Mesh:

    x: np.ndarray
    y: np.ndarray
    meshdf: pd.DataFrame

    @staticmethod
    def from_bounds(bounds, nx=None, ny=None):
        # Get bounding box.
        x0, y0, x1, y1 = bounds

        # Figure out pitch.
        if nx is None and ny is None:
            raise ValueError('Either nx or ny (or both) must be set to an integer')
        elif nx is not None and nx < 1:
            raise ValueError('nx or ny, if set, must be 1 or greater')
        elif ny is not None and ny < 1:
            raise ValueError('nx or ny, if set, must be 1 or greater')
        elif nx is not None and ny is not None:
            xpitch = (x1 - x0) / nx
            ypitch = (y1 - y0) / ny
        elif ny is None:
            xpitch = ypitch = (x1 - x0) / nx
        elif nx is None:
            xpitch = ypitch = (y1 - y0) / ny

        nx = int((x1 - x0) // xpitch)
        margin = (x1 - x0 - nx * xpitch) / 2
        x = np.linspace(x0 + margin, x1 - margin, nx)

        ny = int((y1 - y0) // xpitch)
        margin = (y1 - y0 - ny * xpitch) / 2
        y = np.linspace(y0 + margin, y1 - margin, ny)

        meshx, meshy = np.meshgrid(x, y)
        ix, iy = np.meshgrid(np.arange(len(x)), np.arange(len(y)))

        meshdf = pd.DataFrame().assign(
            x = meshx.ravel(),
            y = meshy.ravel(),
            ix = ix.ravel(),
            iy = iy.ravel())

        return Mesh(x, y, meshdf)

    @staticmethod
    def from_convex_hull(locs, nx=None, ny=None):
        mp = MultiPoint(locs)
        mesh = Mesh.from_bounds(mp.bounds, nx, ny)
        hull = mp.convex_hull
        mask = [hull.contains(Point(p)) for p in mesh.locations()]
        return replace(mesh, meshdf = mesh.meshdf.iloc[mask])

    @staticmethod
    def from_polygon(polygon, nx=None, ny=None):
        mesh = Mesh.from_bounds(polygon.bounds, nx, ny)
        mask = [polygon.contains(Point(p)) for p in mesh.locations()]
        return replace(mesh, meshdf = mesh.meshdf.iloc[mask])

    def locations(self, proj=None):
        loc = self.meshdf[['x', 'y']].values
        if proj is not None:
            loc = (loc @ np.eye(2, 3) + [0, 0, 1]) @ proj
        return loc

    def slice(self, values):
        meshx, meshy = np.meshgrid(self.x, self.y)
        out = np.full([len(self.y), len(self.x)], float('nan'))
        out[(self.meshdf['iy'], self.meshdf['ix'])] = values
        return meshx, meshy, out



from .model import *
from .krige import *
from .mesh import *
from .gp import Observation

__version__ = '0.6.6'

__all__ = []
__all__.extend(model.__all__)
__all__.extend(krige.__all__)
__all__.extend(mesh.__all__)
__all__.extend('Observation')



import numpy as np
from geostat import gp, Model, NormalizingFeaturizer


def test_gp2d():
    # Create 100 random locations in a square centered on the origin.
    locs1 = np.random.uniform(-1., 1., [500, 2])

    # Initialize featurizer of location for trends.
    def trend_terms(x, y): return x, y, x*y
    featurizer = NormalizingFeaturizer(trend_terms, locs1)
    covariance = gp.TrendPrior(featurizer) + gp.SquaredExponential(sill=1.) + gp.Noise()

    # Generating GP.
    model1 = Model(
        latent = covariance,
        parameters = dict(alpha=1., range=0.5, nugget=1.),
        verbose=True)

    # Generate data.
    vals1 = model1.generate(locs1).vals

    # Fit GP.
    model2 = Model(
        latent = covariance,
        parameters = dict(alpha=2., range=1., nugget=0.5),
        verbose=True).fit(locs1, vals1, iters=200)

    # Interpolate using GP.
    N = 20
    xx, yy = np.meshgrid(np.linspace(-1, 1, N), np.linspace(-1, 1, N))
    locs2 = np.stack([xx, yy], axis=-1).reshape([-1, 2])

    mean, var = model2.predict(locs2)
    mean2, var2 = model2.predict(locs2)

    assert np.all(mean == mean2)
    assert np.all(var == var2)


def test_gp3d():
    # Create 100 random locations in a square centered on the origin.
    locs1 = np.random.uniform(-1., 1., [500, 3])

    # Initialize featurizer of location for trends.
    def trend_terms(x, y, z): return z, z*z
    featurizer = NormalizingFeaturizer(trend_terms, locs1)
    covariance = \
        gp.TrendPrior(featurizer) + \
        gp.GammaExponential(scale=[1., 1., 'zscale']) + \
        gp.Delta(axes=[0, 1]) + \
        gp.Noise()

    # Generating GP.
    model1 = Model(
        latent = covariance,
        parameters = dict(alpha=1., zscale=5., range=0.5, sill=1., gamma=1., dsill=1., nugget=1.),
        verbose=True)

    # Generate data.
    vals1 = model1.generate(locs1).vals

    # Fit GP.
    model2 = Model(
        latent = covariance,
        parameters = dict(alpha=2., zscale=1., range=1.0, sill=0.5, gamma=0.5, dsill=0.5, nugget=0.5),
        verbose=True).fit(locs1, vals1, iters=500)

    # Interpolate using GP.
    N = 10
    xx, yy, zz = np.meshgrid(np.linspace(-1, 1, N), np.linspace(-1, 1, N), np.linspace(-1, 1, N))
    locs2 = np.stack([xx, yy, zz], axis=-1).reshape([-1, 3])

    mean, var = model2.predict(locs2)
    mean2, var2 = model2.predict(locs2)

    assert np.all(mean == mean2)
    assert np.all(var == var2)


def test_gp3d_stacked():
    # Create random locations centered on the origin.
    locs1 = np.random.normal(size=[1000, 3])

    # Initialize featurizer of location for trends.
    def trend_terms(x, y, z): return z, z*z
    featurizer = NormalizingFeaturizer(trend_terms, locs1)

    # Covariance structure
    covariance = \
        gp.TrendPrior(featurizer) + \
        gp.SquaredExponential(range='r1', sill='s1', scale=[1., 1., 'zscale']) + \
        gp.SquaredExponential(range='r2', sill='s2', scale=[1., 1., 0.]) + \
        gp.Noise()

    # Generating GP.
    model1 = Model(
        latent = covariance,
        parameters = dict(alpha=1., zscale=5., r1=0.25, s1=1., r2=1.0, s2=0.25, nugget=1.),
        verbose=True)

    # Generate data.
    vals1 = model1.generate(locs1).vals

    # Fit GP.
    model2 = Model(
        latent = covariance,
        parameters = dict(alpha=2., zscale=2.5, r1=0.125, s1=0.5, r2=0.5, s2=0.125, nugget=0.5),
        verbose=True).fit(locs1, vals1, iters=500)

    # Interpolate using GP.
    N = 10
    xx, yy, zz = np.meshgrid(np.linspace(-1, 1, N), np.linspace(-1, 1, N), np.linspace(-1, 1, N))
    locs2 = np.stack([xx, yy, zz], axis=-1).reshape([-1, 3])

    mean, var = model2.predict(locs2)
    mean2, var2 = model2.predict(locs2)

    assert np.all(mean == mean2)
    assert np.all(var == var2)
